/*
 *   linux/dwcs/dwcs.c
 *
 *   Copyright (C) 1999 Rich West
 *
 */

/* 
 *   DWCS kernel-loadable module.
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/utsname.h>
#include <linux/tqueue.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <sys/syscall.h>
#include <asm/segment.h>
#include <errno.h>

#include <linux/mm.h>
#include <linux/kernel_stat.h>
#include <linux/fdreg.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/smp_lock.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/mmu_context.h>
#include <asm/semaphore-helper.h>
#include <asm/system.h>

#include <linux/timex.h>

#define idle_task (&init_task)
#define can_schedule(p) (1)
#define LIMIT (PAGE_SIZE-80)   /* Set limit on amount of data in /proc/dwcs. */

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);
void my_printk(char *string);
void (*_dummy_load_scheduler_fn)(void);
void (*_dummy_unload_scheduler_fn)(void);
void (*_dummy_DWCS_scheduler_fn)(void);

asmlinkage void sys_DWCS_load(void); 
asmlinkage void sys_DWCS_unload(void); 
asmlinkage void DWCS_schedule(void); 

extern void *sys_call_table[];
extern int DWCS_module_loaded;
extern struct task_struct *prev_dwcs_task;
extern struct task_struct *next_dwcs_task;
extern int nr_running;
extern unsigned long volatile jiffies_sim;

#ifdef SIM_MODE
#define JIFFIES jiffies_sim
#else
#define JIFFIES jiffies
#endif

#define DWCS_SERVICED_THIS_PERIOD 0x10000000

int proc_read_dwcs (char *buf, char **start, off_t offset,
		    int len, int unused)
{

#ifdef PROC_DWCS
  struct task_struct * p = init_task.next_run;

  len = 0;

  /* Display runqueue statistics. */
  len += sprintf (buf+len, "Jiffies=%lu Current=%d Prev=%d Next=%d\n",
		  JIFFIES, 
		  current->pid, 
		  current->prev_run->pid,
		  current->next_run->pid);
  
  len += sprintf (buf+len, "  pid name     deadline   period     own   "
		  "owd   cwn   cwd   viols      late       on-time    "
		  "flags srv-time(x10mS)\n");
  while (p != &init_task) {
  /*for_each_task (p) {*/
    if (len > LIMIT)
      return len;
    
    len += sprintf (buf+len, 
		    "%5d %-8.8s %-10ld %-10ld %-5d %-5d %-5d %-5d "
		    "%-10lu %-10lu %-10lu %-5x %-8lu\n",
		    p->pid,
		    p->comm,
		    p->dwcs_params->deadline,
		    p->dwcs_params->period,
		    p->dwcs_params->own,
		    p->dwcs_params->owd,
		    p->dwcs_params->cwn,
		    p->dwcs_params->cwd,
		    p->dwcs_params->violations,
		    p->dwcs_params->late,
		    p->dwcs_params->on_time,
		    p->dwcs_params->flags,
		    p->dwcs_params->service_time);
    
    p = p->next_run;
  }
  return len;
#else
  return 0;
#endif  

}



/* Create an entry in /proc for accessing DWCS state information. */
struct proc_dir_entry proc_dwcs = {
        0,                 /* low_ino: the inode -- dynamic */
        4, "dwcs",         /* len of name and name */
        S_IFREG | S_IRUGO, /* mode */
        1, 0, 0,           /* nlinks, owner, group */
        0, NULL,           /* size - unused; operations -- use default */
        &proc_read_dwcs,   /* function used to read data */
        /* nothing more */
    };


/* Initialize module. */
int init_module(void) 
{
  /* Store original dummy functions for system calls. */
  _dummy_load_scheduler_fn     = sys_call_table[__NR_load_scheduler];
  _dummy_unload_scheduler_fn   = sys_call_table[__NR_unload_scheduler];
  _dummy_DWCS_scheduler_fn     = sys_call_table[__NR_DWCS_scheduler];
  
  /* Redirect  syscall table entries to the syscall fns in this module. */
  sys_call_table[__NR_load_scheduler]     = sys_DWCS_load;
  sys_call_table[__NR_unload_scheduler]   = sys_DWCS_unload;
  sys_call_table[__NR_DWCS_scheduler]     = DWCS_schedule;

  proc_register(&proc_root, &proc_dwcs);
  
  return 0;
}


/* Remove module. */
void cleanup_module(void) {

  /* Restore original dummy functions for system calls. */
  sys_call_table[__NR_load_scheduler]     = _dummy_load_scheduler_fn;
  sys_call_table[__NR_unload_scheduler]   = _dummy_unload_scheduler_fn;
  sys_call_table[__NR_DWCS_scheduler]     = _dummy_DWCS_scheduler_fn;

  /* Reset global flag to tell kernel that DWCS is unloaded. */
  if (DWCS_module_loaded)
    printk (KERN_DEBUG "Disabling DWCS!\n");
  DWCS_module_loaded = 0;

  proc_unregister(&proc_root, proc_dwcs.low_ino);
}


void sys_DWCS_load (void) {

  /* Set global flag to tell kernel that DWCS is loaded. */
  DWCS_module_loaded = 1;

  printk (KERN_DEBUG "Enabling DWCS!\n");
}


void sys_DWCS_unload (void) {

  /* Reset global flag to tell kernel that DWCS is unloaded. */
  DWCS_module_loaded = 0;

  printk (KERN_DEBUG "Disabling DWCS!\n");
}


/*
 * Careful!
 *
 * This has to add the process to the _beginning_ of the
 * run-queue, not the end. See the comment about "This is
 * subtle" in the scheduler proper..
 */
static inline void add_to_runqueue(struct task_struct * p)
{
	struct task_struct *next = init_task.next_run;

	p->prev_run = &init_task;
	init_task.next_run = p;
	p->next_run = next;
	next->prev_run = p;
	nr_running++;
}

static inline void del_from_runqueue(struct task_struct * p)
{
	struct task_struct *next = p->next_run;
	struct task_struct *prev = p->prev_run;

	nr_running--;
	next->prev_run = prev;
	prev->next_run = next;
	p->next_run = NULL;
	p->prev_run = NULL;
}

static inline void move_last_runqueue(struct task_struct * p)
{
	struct task_struct *next = p->next_run;
	struct task_struct *prev = p->prev_run;

	/* remove from list */
	next->prev_run = prev;
	prev->next_run = next;
	/* add back to list */
	p->next_run = &init_task;
	prev = init_task.prev_run;
	init_task.prev_run = p;
	p->prev_run = prev;
	prev->next_run = p;
}

static inline void move_first_runqueue(struct task_struct * p)
{
	struct task_struct *next = p->next_run;
	struct task_struct *prev = p->prev_run;

	/* remove from list */
	next->prev_run = prev;
	prev->next_run = next;
	/* add back to list */
	p->prev_run = &init_task;
	next = init_task.next_run;
	init_task.next_run = p;
	p->next_run = next;
	next->prev_run = p;
}

/*
 * This is the function that decides how desirable a process is..
 * You can weigh different processes against each other depending
 * on what CPU they've run on lately etc to try to handle cache
 * and TLB miss penalties.
 *
 * Return values:
 *	 -1000: never select this
 *	     0: out of time, recalculate counters (but it might still be
 *		selected)
 *	   +ve: "goodness" value (the larger, the better)
 *	 +1000: realtime process, select this.
 */
static inline int goodness(struct task_struct * p, 
			   struct task_struct * prev, 
			   int this_cpu)
{
	int policy = p->policy;
	int weight;

	if (policy & SCHED_YIELD) {
		p->policy = policy & ~SCHED_YIELD;
		return 0;
	}

	/*
	 * Realtime process, select the first one on the
	 * runqueue (taking priorities within processes
	 * into account).
	 */
	if (policy != SCHED_OTHER)
		return 1000 + p->rt_priority;

	/*
	 * Give the process a first-approximation goodness value
	 * according to the number of clock-ticks it has left.
	 *
	 * Don't do any other calculations if the time slice is
	 * over..
	 */
	weight = p->counter;
	if (weight) {
			
#ifdef __SMP__
		/* Give a largish advantage to the same processor...   */
		/* (this is equivalent to penalizing other processors) */
		if (p->processor == this_cpu)
			weight += PROC_CHANGE_PENALTY;
#endif

		/* .. and a slight advantage to the current thread */
		if (p->mm == prev->mm)
			weight += 1;
		weight += p->priority;
	}

	return weight;
}


static void update_service_constraints (struct task_struct * p,
					int check_deadlines) {
  
  if (check_deadlines) {
    while (JIFFIES > p->dwcs_params->deadline) {
      /* printk (KERN_DEBUG "(%d) Deadline missed!\n",
	 p->dwcs_params->cd); */
      
#ifdef PROC_DWCS
      /* Statistics. */
      
      p->dwcs_params->late++;	
      p->dwcs_params->wlate++;
      if (((p->dwcs_params->on_time + 
	    p->dwcs_params->late) %
	   p->dwcs_params->owd) == 0) {
	
	if (p->dwcs_params->wlate > p->dwcs_params->own) {
	  p->dwcs_params->violations++;
	  /*printk (KERN_DEBUG "(%d) VIOLATION!\n",
	    p->dwcs_params->cd);*/
	}
	p->dwcs_params->wlate = 0;
      }
#endif
      
      p->dwcs_params->deadline += p->dwcs_params->period;
      
      if (p->dwcs_params->cwn > 0) {
	p->dwcs_params->cwn--;
	p->dwcs_params->cwd--;
	if ((p->dwcs_params->cwd == 0) && (p->dwcs_params->cwn == 0)) {
	  p->dwcs_params->cwn = p->dwcs_params->own;
	  p->dwcs_params->cwd = p->dwcs_params->owd;
	}
      }
      else { /* p->dwcs_params->cwn == 0. */ 
	if (p->dwcs_params->owd > 0) { 
	  /* 
	   * NOTE: This is not pure EDF if we got here, 
	   * since if it were, x/y = 0/0 and, therefore,
           * p->dwcs_params->owd = 0.
           *
	   * Now, we have a window-constraint violation.
	   * These next instructions will eventually 
           * avoid starvation of p.
	   */
	  p->dwcs_params->cwd++;       
	  p->dwcs_params->flags |= DWCS_VIOLATION;
	}
      }
    }
  }
  
  else {

    /* Service time (and time serviced) is at the resolution of
     * jiffies right now. That is, it is currently some multiple the
     * clock interrupt resolution (10mS on Intels). 
     */
    p->dwcs_params->time_serviced++; 

    /* Adjust service attributes only when a process is serviced in
       its entirety "on-time". */
    if (p->dwcs_params->time_serviced == p->dwcs_params->service_time) {

      if ((p->dwcs_params->cwd > p->dwcs_params->cwn) && 
	  (p->dwcs_params->deadline  != (unsigned long)-1))
	p->dwcs_params->cwd--;
      
      if (((p->dwcs_params->cwd == 0) && (p->dwcs_params->cwn == 0)) ||
	  (p->dwcs_params->flags & DWCS_VIOLATION)) {
	p->dwcs_params->cwn = p->dwcs_params->own;
	p->dwcs_params->cwd = p->dwcs_params->owd;
	p->dwcs_params->flags &= ~DWCS_VIOLATION;
      }
    }

    /* Disable any possible servicing of this task until its next
       request period if it does not support work-conservation. */
    if ((!(p->dwcs_params->flags & DWCS_WORK_CONS)) &&
	(p->dwcs_params->time_serviced == p->dwcs_params->service_time))
      p->dwcs_params->ready_time += p->dwcs_params->period;
    else
      if (p->dwcs_params->deadline != (unsigned long)-1) {
	if (p->dwcs_params->time_serviced == p->dwcs_params->service_time) {
	  p->dwcs_params->flags |= DWCS_SERVICED_THIS_PERIOD;

	  /* The following line tells us when it's safe to
	     reset the DWCS_SERVICED_THIS_PERIOD flag. */
	  p->dwcs_params->ready_time += p->dwcs_params->period;
	}
      }
    
    if (p->dwcs_params->time_serviced == p->dwcs_params->service_time) {
      if (p->dwcs_params->deadline != (unsigned long)-1)
	p->dwcs_params->deadline += p->dwcs_params->period;          
      
      p->dwcs_params->time_serviced = 0;
    }

#ifdef PROC_DWCS
    /* Statistics. */
    p->dwcs_params->on_time++;
    
    if (((p->dwcs_params->on_time + 
	  p->dwcs_params->late) %
	 p->dwcs_params->owd) == 0) {
      
      if (p->dwcs_params->wlate > p->dwcs_params->own) {
	p->dwcs_params->violations++;
	/*printk (KERN_DEBUG "(%d) VIOLATION!\n",
	  p->dwcs_params->cd);*/
      }
      p->dwcs_params->wlate = 0;
    }
#endif
    
  }
  p->dwcs_params->cwc = (p->dwcs_params->cwd == 0) ? 0.0 : 
    (float)p->dwcs_params->cwn/p->dwcs_params->cwd;
}


struct task_struct *
compare_service_constraints (struct task_struct * p, 
			     struct task_struct * next,
			     struct task_struct * prev,
			     int weight_of_p) {

  if ((next->dwcs_params->flags & DWCS_NON_PREEMPTIVE) &&
      (next->dwcs_params->time_serviced < next->dwcs_params->service_time) &&
      (next->dwcs_params->time_serviced > 0))
    return next;

  if ((p->dwcs_params->ready_time > JIFFIES) &&
      (!(p->dwcs_params->flags & DWCS_WORK_CONS)))
    return next;


  /* 
   * Search run queue for highest priority i.e. lowest
   * window-constrained, EDF task and set to `next' task for
   * service in the run queue. 
   */

  if ((p->dwcs_params->deadline == next->dwcs_params->deadline) &&
      (p->dwcs_params->deadline == (unsigned long)-1) &&
      (p->dwcs_params->cwn == next->dwcs_params->cwn) &&
      (p->dwcs_params->cwd == next->dwcs_params->cwd)) {

    int this_cpu = prev->processor;
    int weight_of_next = goodness(next, prev, this_cpu);
   
    /* 
     * NOTE: For non-time-constrained tasks.
     *
     * We could've decided not do deal with `weights'. In that
     * case, a non-blocking process could hog the CPU even if
     * other processes are in the runqueue and have the same DWCS
     * service constraints. To be more `fair' even though we're 
     * supposed to be real-time, we let the process that has the
     * largest weight have the CPU if two processes have the same DWCS
     * service constraints. This behaviour is in keeping with
     * the default Linux scheduler's behaviour. 
     */
    if (weight_of_p > weight_of_next)
      next = p;
  }
  else {
    
    if ((p->dwcs_params->deadline < next->dwcs_params->deadline) ||
	((p->dwcs_params->deadline == next->dwcs_params->deadline) &&
	 (p->dwcs_params->cwc < next->dwcs_params->cwc)) ||
	((p->dwcs_params->deadline == next->dwcs_params->deadline) &&
	 (p->dwcs_params->cwc == 0.0) && (next->dwcs_params->cwc == 0.0) &&
	 (p->dwcs_params->cwd > next->dwcs_params->cwd)) ||
	((p->dwcs_params->deadline == next->dwcs_params->deadline) &&
	 (p->dwcs_params->cwc == next->dwcs_params->cwc) &&
	 (p->dwcs_params->cwc > 0.0) &&
	 (p->dwcs_params->cwn < next->dwcs_params->cwn)) ||
	((p->dwcs_params->deadline != (unsigned long)-1) &&
	 (!(p->dwcs_params->flags & DWCS_SERVICED_THIS_PERIOD)) &&
	 (next->dwcs_params->flags & DWCS_SERVICED_THIS_PERIOD))) {
      
      next = p;
    }
  }
  
  return next;
}


/* The actual DWCS scheduler. */
asmlinkage void DWCS_schedule(void) {

  struct task_struct * prev, * next;
  int this_cpu;

  
  jiffies_sim++;


  run_task_queue(&tq_scheduler);
  
  prev = current;
  this_cpu = prev->processor;
  
  if (in_interrupt())
    goto scheduling_in_interrupt;
  release_kernel_lock(prev, this_cpu);
  
  /* Do "administrative" work here while we don't hold any locks */
  if (bh_active & bh_mask)
    do_bottom_half();
  
  spin_lock(&scheduler_lock);
  spin_lock_irq(&runqueue_lock);
  
  /* move an exhausted RR process to be last.. */
  prev->need_resched = 0;
  
  if (!prev->counter && prev->policy == SCHED_RR) {
    prev->counter = prev->priority;
    move_last_runqueue(prev);
  }
  
  switch (prev->state) {
  case TASK_INTERRUPTIBLE:
    if (signal_pending(prev)) {
      prev->state = TASK_RUNNING;
      break;
    }
  default:
    del_from_runqueue(prev);
  case TASK_RUNNING:
  }
  
  /* this is the scheduler proper: */
  {
    struct task_struct * p = init_task.next_run;
    int c = -1000;

    while (p != &init_task) {
      if (can_schedule(p))
	update_service_constraints(p, 1);
      p = p->next_run;
    }
    p = init_task.next_run;
    
    /* Default process to select.. */
    next = idle_task;
    if (prev->state == TASK_RUNNING) {
      c = goodness(prev, prev, this_cpu);
      if (prev->dwcs_params->ready_time <= JIFFIES) {
	
	/* Reset DWCS_SERVICED_THIS PERIOD in case it was set. */
	prev->dwcs_params->flags &= ~DWCS_SERVICED_THIS_PERIOD;
	
	next = prev;
      }
    }

    /*
     * This is subtle.
     * Note how we can enable interrupts here, even
     * though interrupts can add processes to the run-
     * queue. This is because any new processes will
     * be added to the front of the queue, so "p" above
     * is a safe starting point.
     * run-queue deletion and re-ordering is protected by
     * the scheduler lock
     */
    spin_unlock_irq(&runqueue_lock);
    /*
     * Note! there may appear new tasks on the run-queue during this, as
     * interrupts are enabled. However, they will be put on front of the
     * list, so our list starting at "p" is essentially fixed.
     */
    
    while (p != &init_task) {
      if (can_schedule(p)) {
	int weight = goodness(p, prev, this_cpu);
	if (weight > c)
	  c = weight;
	next = compare_service_constraints(p, next, prev, weight);
      }
      p = p->next_run;
    }
    
    /* Do we need to re-calculate counters? */
    if (!c) {
      struct task_struct *p;
      read_lock(&tasklist_lock);
      for_each_task(p)
	p->counter = (p->counter >> 1) + p->priority;
      read_unlock(&tasklist_lock);
    }
  }
  
  /*
   * maintain the per-process 'average timeslice' value.
   * (this has to be recalculated even if we reschedule to
   * the same process) Currently this is only used on SMP:
   */
#ifdef __SMP__
  {
    cycles_t t, this_slice;
    
    t = get_cycles();
    this_slice = t - sched_data->last_schedule;
    sched_data->last_schedule = t;
    
    /*
     * Simple, exponentially fading average calculation:
     */
    prev->avg_slice = this_slice + prev->avg_slice;
    prev->avg_slice >>= 1;
  }
  
  /*
   * We drop the scheduler lock early (it's a global spinlock),
   * thus we have to lock the previous process from getting
   * rescheduled during switch_to().
   */
  next->processor = this_cpu;
  next->has_cpu = 1;
  spin_unlock(&scheduler_lock);
#endif /* __SMP__ */

  update_service_constraints(next, 0);
  prev_dwcs_task = prev;
  next_dwcs_task = next;
  
  return;
  
 scheduling_in_interrupt:
  printk("Scheduling in interrupt\n");
  *(int *)0 = 0;
}
